#include <nemo/stdint.h>
#include <nemo/byte.h>
#include <nemo/open.h>
#include <nemo/error.h>
#include <nemo/direntry.h>
#include <nemo/stralloc.h>
#include <nemo/sa_vector.h>
#include <nemo/openreadclose.h>
#include <nemo/strerr.h>
#include <nemo/pathexec.h>
#include <nemo/unix.h>

#include "die.h"

const char USAGE[] = "dir child";

int main(int argc, char **argv)
{
  static stralloc line = STRALLOC;
  int fd_orig_dir;
  const char *fn;
  DIR *dir;
  direntry *d;

  PROGRAM = argv[0];
  if (argc < 3) die_usage();

  fn = *++argv;
  ++argv;

  fd_orig_dir = open_read(".");
  if (fd_orig_dir < 0) die_chdir("current directory");
  if (chdir(fn) < 0) die_chdir(fn);

  dir = opendir(".");
  if (!dir) die_read(fn);
  for (;;) {
    errno = 0;
    d = readdir(dir);
    if (!d) {
      if (errno) die_read(fn);
      break;
    }
    if (d->d_name[0] == '.') continue;
    if (openreadlineclose(d->d_name, &line) < 0) die_read2(fn, d->d_name);
    if (line.len) {
      stralloc_replace(&line, '\0', '\n');
      if (!stralloc_0(&line)) die_nomem();
      if (!pathexec_env(d->d_name, line.s)) die_nomem();
    }
    else {
      if (!pathexec_env(d->d_name, 0)) die_nomem();
    }
  }
  closedir(dir);

  if (fchdir(fd_orig_dir) < 0) die_chdir("starting directory");
  close(fd_orig_dir);

  pathexec(argv);
  die_exec(*argv);
  return 0;  /* lint */
}
