#include <nemo/stdint.h>
#include <nemo/unixtypes.h>
#include <nemo/error.h>
#include <nemo/strerr.h>
#include <nemo/pathexec.h>
#include <nemo/exit.h>
#include <nemo/unix.h>
#include <nemo/fd.h>
#include <nemo/fork.h>
#include <nemo/wait.h>

#include "die.h"

const char USAGE[] = "child";

static int pid;

int main(int argc, char **argv, char **envp)
{
  unsigned int i;
  int wstat;
  int pi[2];
  ssize_t r;
  char ch;

  PROGRAM = argv[0];
  if (argc < 2) die_usage();

  if (pipe(pi) < 0) die_create_pipe("");

  switch (pid = fork()) {
    case -1:
      die_fork();
      break;
    case 0:
      close(pi[0]);
      for (i = 0; i < 30; ++i) {
        dup(pi[1]);
      }
      pathexec_run(argv[1], argv + 1, envp);
      die_exec(argv[1]);
      break;
    default:
      break;
  }

  close(pi[1]);

  for (;;) {
    r = read(pi[0], &ch, 1);
    if ((r < 0) && (errno == error_intr)) continue;
    if (r == 1) continue;
    break;
  }

  if (wait_pid(&wstat, pid) < 0) die_wait(argv[1]);
  if (wait_crashed(wstat)) die_child_crashed(argv[1]);
  _exit(wait_exitcode(wstat));
}
