#include <nemo/stdint.h>
#include <nemo/unixtypes.h>

#include <nemo/direntry.h>
#include <nemo/alloc.h>
#include <nemo/stralloc.h>
#include <nemo/djbio.h>
#include <nemo/djbiofd.h>
#include <nemo/strerr.h>
#include <nemo/error.h>
#include <nemo/open.h>
#include <nemo/lock.h>
#include <nemo/scan.h>
#include <nemo/str.h>
#include <nemo/byte.h>
#include <nemo/seek.h>
#include <nemo/wait.h>
#include <nemo/coe.h>
#include <nemo/env.h>
#include <nemo/fd.h>
#include <nemo/sig.h>
#include <nemo/exit.h>
#include <nemo/unix.h>
#include <nemo/tai.h>
#include <nemo/taia.h>
#include <nemo/fork.h>
#include <nemo/macro_unused.h>

#include "match.h"
#include "deepsleep.h"
#include "die.h"

/* #include <sys/types.h> */
#include <sys/stat.h>

static const char WARNING[] = "multilog: warning: ";

static const char FN_CURRENT[] = "current";

static void pause2(const char *s1, const char *s2)
{
  strerr_warn4sys(WARNING, s1, s2, ", pausing");
  deepsleep(5);
}

static void pause3(const char *s1, const char *s2, const char *s3)
{
  strerr_warn5sys(WARNING, s1, s2, s3, ", pausing");
  deepsleep(5);
}

static void pause4(const char *s1, const char *s2, const char *s3, const char *s4)
{
  strerr_warn6sys(WARNING, s1, s2, s3, s4, ", pausing");
  deepsleep(5);
}

static int fd_start_dir;

static int *f;

static void f_init(char **script)
{
  unsigned int i;
  int fd;

  for (i = 0; script[i]; ++i)
    ;

  f = (int *) alloc(i * sizeof(*f));
  if (!f) die_nomem();

  for (i = 0; script[i]; ++i) {
    fd = -1;
    if (script[i][0] == '=') {
      if (fchdir(fd_start_dir) < 0) die_chdir("starting directory");
      fd = open_append(script[i] + 1);
      if (fd < 0) die_create(script[i] + 1);
      close(fd);
      fd = open_write(script[i] + 1);
      if (fd < 0) die_write(script[i] + 1);
      coe(fd);
    }
    f[i] = fd;
  }
}

static struct cyclog {
  char buf[512];
  djbio ss;
  int fd_current;
  unsigned long bytes;
  unsigned long num;
  unsigned long size;
  char *processor;
  char *dir;
  int fd_dir;
  int fd_lock;
  unsigned int flag_selected;
} *c;
static unsigned int cnum;

static char fn[40];

static int files_fit(struct cyclog *d)
{
  DIR *dir;
  direntry *x;
  unsigned int count;
  unsigned int i;

  dir = opendir(".");
  if (!dir) return -1;

  fn[0] = '@';
  fn[1] = 'z';
  fn[2] = '\0';

  count = 0;
  for (;;) {
    errno = 0;
    x = readdir(dir);
    if (!x) break;
    if (x->d_name[0] == '@') {
      if (str_len(x->d_name) >= 25) {
        ++count;
        if (str_diff(x->d_name, fn) < 0) {
          for (i = 0;i < 25;++i) {
            fn[i] = x->d_name[i];
          }
          fn[25] = '\0';
        }
      }
    }
  }
  if (errno) {
    closedir(dir);
    return -1;
  }
  closedir(dir);

  if (count < d->num) return 1;

  dir = opendir(".");
  if (!dir) return -1;

  for (;;) {
    errno = 0;
    x = readdir(dir);
    if (!x) break;
    if (x->d_name[0] == '@') {
      if (str_len(x->d_name) >= 25) {
        if (str_start(x->d_name, fn)) {
          unlink(x->d_name);
          break;
        }
      }
    }
  }
  if (errno) {
    closedir(dir);
    return -1;
  }
  closedir(dir);
  return 0;
}

static void finish(struct cyclog *d, const char *file, const char *code)
{
  struct stat st;

  for (;;) {
    if (stat(file, &st) == 0) break;
    if (errno == error_noent) return;
    pause4("unable to stat ", d->dir, "/", file);
  }

  if (st.st_nlink == 1) {
    for (;;) {
      taia_timestamp(fn);
      fn[25] = '.';
      fn[26] = code[0];
      fn[27] = 0;

      if (link(file, fn) == 0) break;
      pause4("unable to link to ", d->dir, "/", fn);
    }
  }

  while (unlink(file) < 0) {
    pause4("unable to remove ", d->dir, "/", file);
  }

  for (;;) {
    switch (files_fit(d)) {
      case 1:
        return;
      case -1:
        pause2("unable to read ", d->dir);
        break;
      default:
        break;
    }
  }
}

static void start_processor(struct cyclog *d)
{
  static char shell[] = "sh";
  static char param1[] = "-c";
  static char *args[4] = { shell, param1, 0, 0 };
  int fd;

  sig_uncatch(sig_term);
  sig_uncatch(sig_alarm);
  sig_unblock(sig_term);
  sig_unblock(sig_alarm);

  fd = open_read("previous");
  if (fd < 0) return;
  if (fd_move(0, fd) < 0) return;
  fd = open_trunc("processed");
  if (fd < 0) return;
  if (fd_move(1, fd) < 0) return;
  fd = open_read("state");
  if (fd < 0) return;
  if (fd_move(4, fd) < 0) return;
  fd = open_trunc("newstate");
  if (fd < 0) return;
  if (fd_move(5, fd) < 0) return;
/*
  args[0] = "sh";
  args[1] = "-c";
*/
  args[2] = d->processor;
/*
  args[3] = 0;
*/
  execve("/bin/sh", args, environ);
}

static void full_current(struct cyclog *d)
{
  int fd;
  int pid;
  int wstat;

  while (fchdir(d->fd_dir) < 0) {
    pause2("unable to switch to ", d->dir);
  }

  while (fsync(d->fd_current) < 0) {
    pause3("unable to write ", d->dir, "/current to disk");
  }
  close(d->fd_current);

  while (rename("current", "previous") < 0) {
    pause2("unable to rename current to previous in directory ", d->dir);
  }
  while ((d->fd_current = open_append("current")) < 0) {
    pause3("unable to create ", d->dir, "/current");
  }
  coe(d->fd_current);
  d->bytes = 0;
  while (fchmod(d->fd_current, 0644) < 0) {
    pause3("unable to set mode of ", d->dir, "/current");
  }

  while (chmod("previous", 0744) < 0) {
    pause3("unable to set mode of ", d->dir, "/previous");
  }

  if (!d->processor) {
    finish(d, "previous", "s");
  }
  else {
    for (;;) {
      while ((pid = fork()) < 0) {
        pause2("unable to fork for processor in ", d->dir);
      }
      if (!pid) {
        start_processor(d);
        die_exec(d->processor);
      }
      if (wait_pid(&wstat, pid) < 0) {
        pause2("wait failed for processor in ", d->dir);
      }
      else if (wait_crashed(wstat)) {
        pause2("processor crashed in ", d->dir);
      }
      else if (!wait_exitcode(wstat)) {
        break;
      }
      pause2("processor failed in ", d->dir);
    }

    while ((fd = open_append("processed")) < 0) {
      pause3("unable to create ", d->dir, "/processed");
    }
    while (fsync(fd) < 0) {
      pause3("unable to write ", d->dir, "/processed to disk");
    }
    while (fchmod(fd, 0744) < 0) {
      pause3("unable to set mode of ", d->dir, "/processed");
    }
    close(fd);

    while ((fd = open_append("newstate")) < 0) {
      pause3("unable to create ", d->dir, "/newstate");
    }
    while (fsync(fd) < 0) {
      pause3("unable to write ", d->dir, "/newstate to disk");
    }
    close(fd);

    while (unlink("previous") < 0) {
      pause3("unable to remove ", d->dir, "/previous");
    }
    while (rename("newstate", "state") < 0) {
      pause2("unable to rename newstate to state in directory ", d->dir);
    }
    finish(d, "processed", "s");
  }
}

static ssize_t c_write(int pos, const void *b, size_t len)
{
  struct cyclog *d;
  unsigned int i;
  ssize_t w;
  const char *buf;

  buf = b;
  d = c + pos;

  if (d->bytes >= d->size) {
    full_current(d);
  }

  if (len >= d->size - d->bytes) {
    len = d->size - d->bytes;
  }

  if (d->bytes + len >= d->size - 2000) {
    i = byte_rchr(buf, (unsigned int)len, '\n');
    if (i < len) {
      len = i + 1;
    }
  }

  for (;;) {
    w = write(d->fd_current, buf, len);
    if (w > 0) break;
    pause3("unable to write to ", d->dir, "/current");
  }

  d->bytes += (unsigned long)w;
  if (d->bytes >= d->size - 2000) {
    if (buf[w - 1] == '\n') {
      full_current(d);
    }
  }

  return w;
}

static void restart(struct cyclog *d)
{
  struct stat st;
  unsigned int flag_processed;
  int fd;

  if (fchdir(fd_start_dir) < 0) die_chdir("starting directory");

  mkdir(d->dir, 0700);
  d->fd_dir = open_read(d->dir);
  if ((d->fd_dir < 0) || (fchdir(d->fd_dir) < 0)) die_open(d->dir);
  coe(d->fd_dir);

  d->fd_lock = open_append("lock");
  if ((d->fd_lock < 0) || (lock_exnb(d->fd_lock) < 0)) die_lock(d->dir);
  coe(d->fd_lock);

  if (stat(FN_CURRENT, &st) < 0) {
    if (errno != error_noent) die_stat(d->dir, FN_CURRENT);
  }
  else {
    if (st.st_mode & 0100) {
      fd = open_append(FN_CURRENT);
      if (fd < 0) die_append(d->dir, FN_CURRENT);
      if (fchmod(fd, 0644) < 0) die_chmod(d->dir, FN_CURRENT);
      coe(fd);
      d->fd_current = fd;
      d->bytes = (unsigned long)st.st_size;
      return;
    }
  }

  unlink("state");
  unlink("newstate");

  flag_processed = 0;
  if (stat("processed", &st) < 0) {
    if (errno != error_noent) die_stat(d->dir, "processed");
  }
  else if (st.st_mode & 0100) {
    flag_processed = 1;
  }

  if (flag_processed) {
    unlink("previous");
    finish(d, "processed", "s");
  }
  else {
    unlink("processed");
    finish(d, "previous", "u");
  }

  finish(d, FN_CURRENT, "u");

  fd = open_trunc("state");
  if (fd < 0) die_write2(d->dir, "state");
  close(fd);

  fd = open_append(FN_CURRENT);
  if (fd < 0) die_write2(d->dir, FN_CURRENT);
  if (fchmod(fd, 0644) < 0) die_chmod(d->dir, FN_CURRENT);
  coe(fd);
  d->fd_current = fd;
  d->bytes = 0;
}

static void c_init(char **script)
{
  unsigned int i;
  struct cyclog *d;
  char *processor;
  unsigned long num;
  unsigned long size;

  cnum = 0;
  for (i = 0; script[i]; ++i) {
    if ((script[i][0] == '.') || (script[i][0] == '/')) {
      ++cnum;
    }
  }

  c = (struct cyclog *) alloc(cnum * sizeof(*c));
  if (!c) die_nomem();

  d = c;
  processor = 0;
  num = 10;
  size = 99999;

  for (i = 0; script[i]; ++i) {
    if (script[i][0] == 's') {
      scan_ulong(script[i] + 1, &size);
      if (size < 4096) {
        size = 4096;
      }
      else if (size > 16777215) {
        size = 16777215;
      }
    }
    else if (script[i][0] == 'n') {
      scan_ulong(script[i] + 1, &num);
      if (num < 2) {
        num = 2;
      }
    }
    else if (script[i][0] == '!') {
      processor = script[i] + 1;
    }
    else if ((script[i][0] == '.') || (script[i][0] == '/')) {
      d->num = num;
      d->size = size;
      d->processor = processor;
      d->dir = script[i];
      djbio_initwrite(&d->ss, c_write, (int)(d - c), d->buf, sizeof d->buf);
      restart(d);
      ++d;
    }
  }
}

static void c_quit(void)
{
  unsigned int j;

  for (j = 0; j < cnum; ++j) {
    djbio_flush(&c[j].ss);
    while (fsync(c[j].fd_current) < 0) {
      pause3("unable to write ", c[j].dir, "/current to disk");
    }
    while (fchmod(c[j].fd_current, 0744) < 0) {
      pause3("unable to set mode of ", c[j].dir, "/current");
    }
  }
}

static unsigned int flag_exit_asap = 0;
static unsigned int flag_force_rotate = 0;
static unsigned int flag_new_line = 1;

static void exit_asap(int i __UNUSED__)
{
  flag_exit_asap = 1;
}

static void force_rotate(int i __UNUSED__)
{
  flag_force_rotate = 1;
}

static ssize_t flushread(int fd, void *buf, size_t len)
{
  unsigned int j;
  char *x;
  ssize_t rlen;

  for (j = 0; j < cnum; ++j) {
    djbio_flush(&c[j].ss);
  }
  if (flag_force_rotate) {
    for (j = 0;j < cnum;++j) {
      if (c[j].bytes > 0) {
        full_current(&c[j]);
      }
    }
    flag_force_rotate = 0;
  }

  if (!len) return 0;
  if (flag_exit_asap) {
    if (flag_new_line) return 0;
    len = 1;
  }

  sig_unblock(sig_term);
  sig_unblock(sig_alarm);

  rlen = read(fd, buf, len);

  sig_block(sig_term);
  sig_block(sig_alarm);

  if (rlen <= 0) {
    return rlen;
  }
  x = buf;
  flag_new_line = (x[rlen - 1] == '\n');
  return rlen;
}

static char inbuf[1024];
static djbio io_in = DJBIO_INITREAD(flushread, 0, inbuf, sizeof inbuf);

static char line[1001];
static unsigned int line_len; /* 0 <= line_len <= 1000 */

static void doit(char **script)
{
  char *action;
  unsigned int flag_eof;
  unsigned int i;
  unsigned int j;
  unsigned int flag_selected;
  unsigned int flag_timestamp;
  byte_t ch;

  flag_timestamp = 0;
  if (script[0]) {
    if (script[0][0] == 't') {
      flag_timestamp = 1;
    }
  }

  for (i = 0; i <= 1000; ++i) {
    line[i] = '\n';
  }
  line_len = 0;

  flag_eof = 0;
  for (;;) {
    for (i = 0; i < line_len; ++i) {
      line[i] = '\n';
    }
    line_len = 0;

    while (line_len < 1000) {
      if (djbio_GETC(&io_in, &ch) <= 0) {
        if (!line_len) return;
        flag_eof = 1;
        break;
      }
      if (!line_len) {
        if (flag_timestamp) {
          taia_timestamp(line);
          line[25] = ' ';
          line_len = 26;
        }
      }
      if (ch == '\n') {
        break;
      }
      line[line_len++] = (char)ch;
    }

    flag_selected = 1;
    j = 0;
    for (i = 0; (action = script[i]); ++i) {
      switch (*action) {
        case '+':
          if (!flag_selected) {
            if (match(action + 1, line, line_len)) {
              flag_selected = 1;
            }
          }
          break;
        case '-':
          if (flag_selected) {
            if (match(action + 1, line, line_len)) {
              flag_selected = 0;
            }
          }
          break;
        case 'e':
          if (flag_selected) {
            if (line_len > 200) {
              djbio_put(djbiofd_err, line, 200);
              djbio_puts(djbiofd_err, "...\n");
            }
            else {
              djbio_put(djbiofd_err, line, line_len);
              djbio_puts(djbiofd_err, "\n");
            }
            djbio_flush(djbiofd_err);
          }
          break;
        case '=':
          if (flag_selected) {
            for (;;) {
              while (seek_begin(f[i]) < 0) {
                pause2("unable to move to beginning of ", action + 1);
              }
              if (write(f[i], line, 1001) == 1001) {
                break;
              }
              pause2("unable to write ", action + 1);
            }
          }
          break;
        case '.':
        case '/':
          c[j].flag_selected = flag_selected;
          ++j;
          break;
        default:
          break;
      }
    }

    for (j = 0; j < cnum; ++j) {
      if (c[j].flag_selected) {
        djbio_put(&c[j].ss, line, line_len);
      }
    }

    if (line_len == 1000) {
      for (;;) {
        if (djbio_GETC(&io_in, &ch) <= 0) {
          flag_eof = 1;
          break;
        }
        if (ch == '\n') {
          break;
        }
        for (j = 0; j < cnum; ++j) {
          if (c[j].flag_selected) {
            djbio_PUTC(&c[j].ss, ch);
          }
        }
      }
    }

    for (j = 0; j < cnum; ++j) {
      if (c[j].flag_selected) {
        ch = '\n';
        djbio_PUTC(&c[j].ss, ch);
      }
    }

    if (flag_eof) return;
  }
}

int main(int argc, char **argv)
{
  PROGRAM = argv[0];
  umask(022);

  fd_start_dir = open_read(".");
  if (fd_start_dir < 0) die_chdir("current directory");
  coe(fd_start_dir);

  sig_block(sig_term);
  sig_block(sig_alarm);
  sig_catch(sig_term, exit_asap);
  sig_catch(sig_alarm, force_rotate);

  if (!argc) _exit(111);  /* lint (argc) */
  ++argv;
  f_init(argv);
  c_init(argv);
  doit(argv);
  c_quit();
  _exit(0);
}
