#include <sys/types.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <nemo/stdint.h>
#include <nemo/uint32.h>
#include <nemo/pathexec.h>
#include <nemo/sgetopt.h>
#include <nemo/strerr.h>
#include <nemo/scan.h>
#include <nemo/str.h>
#include <nemo/env.h>

#include "die.h"

const char USAGE[] = "[-a allbytes] [-c corebytes] [-D | -d databytes] [-f filebytes] [-l lockbytes] [-m membytes] [-O | -o openfiles] [-P | -p processes] [-r residentbytes] [-S | -s stackbytes] [-T | -t cpusecs] child";

static void doit(int resource, const char *arg)
{
  unsigned long u;
  struct rlimit r;

  if (getrlimit(resource, &r) < 0) die_sys("getrlimit failed");

  if (str_equal(arg, "=")) {
    r.rlim_cur = r.rlim_max;
  }
  else {
    if (arg[scan_ulong(arg, &u)]) die_usage();
    r.rlim_cur = (rlim_t)u;
    if (r.rlim_cur > r.rlim_max) {
      r.rlim_cur = r.rlim_max;
    }
  }

  if (setrlimit(resource, &r) < 0) die_sys("setrlimit failed");
}

int main(int argc, char **argv, char **envp)
{
  int opt;
  const char *x;

  PROGRAM = argv[0];
  while ((opt = getopt(argc, argv, "DOPSTa:c:d:f:l:m:o:p:r:s:t:")) != opteof) {
    switch (opt) {
      case 'a':
#ifdef RLIMIT_AS
        doit(RLIMIT_AS, optarg);
#endif
#ifdef RLIMIT_VMEM
        doit(RLIMIT_VMEM, optarg);
#endif
        break;
      case 'c':
#ifdef RLIMIT_CORE
        doit(RLIMIT_CORE, optarg);
#endif
        break;
      case 'd':
#ifdef RLIMIT_DATA
        doit(RLIMIT_DATA, optarg);
#endif
        break;
      case 'D':
        x = env_get("DATALIMIT");
        if (!x) die_env("DATALIMIT");
#ifdef RLIMIT_DATA
        doit(RLIMIT_DATA, x);
#endif
        break;
      case 'f':
#ifdef RLIMIT_FSIZE
        doit(RLIMIT_FSIZE, optarg);
#endif
        break;
      case 'l':
#ifdef RLIMIT_MEMLOCK
        doit(RLIMIT_MEMLOCK, optarg);
#endif
        break;
      case 'm':
#ifdef RLIMIT_DATA
        doit(RLIMIT_DATA, optarg);
#endif
#ifdef RLIMIT_STACK
        doit(RLIMIT_STACK, optarg);
#endif
#ifdef RLIMIT_MEMLOCK
        doit(RLIMIT_MEMLOCK, optarg);
#endif
#ifdef RLIMIT_VMEM
        doit(RLIMIT_VMEM, optarg);
#endif
#ifdef RLIMIT_AS
        doit(RLIMIT_AS, optarg);
#endif
        break;
      case 'o':
#ifdef RLIMIT_NOFILE
        doit(RLIMIT_NOFILE, optarg);
#endif
#ifdef RLIMIT_OFILE
        doit(RLIMIT_OFILE, optarg);
#endif
        break;
      case 'O':
        x = env_get("FDLIMIT");
        if (!x) die_env("FDLIMIT");
#ifdef RLIMIT_NOFILE
        doit(RLIMIT_NOFILE, x);
#endif
#ifdef RLIMIT_OFILE
        doit(RLIMIT_OFILE, x);
#endif
        break;
      case 'p':
#ifdef RLIMIT_NPROC
        doit(RLIMIT_NPROC, optarg);
#endif
        break;
      case 'P':
        x = env_get("PROCESSLIMIT");
        if (!x) die_env("PROCESSLIMIT");
#ifdef RLIMIT_NPROC
        doit(RLIMIT_NPROC, x);
#endif
        break;
      case 'r':
#ifdef RLIMIT_RSS
        doit(RLIMIT_RSS, optarg);
#endif
        break;
      case 's':
#ifdef RLIMIT_STACK
        doit(RLIMIT_STACK, optarg);
#endif
        break;
      case 'S':
        x = env_get("STACKLIMIT");
        if (!x) die_env("STACKLIMIT");
#ifdef RLIMIT_STACK
        doit(RLIMIT_STACK, x);
#endif
        break;
      case 't':
#ifdef RLIMIT_CPU
        doit(RLIMIT_CPU, optarg);
#endif
        break;
      case 'T':
        x = env_get("CPULIMIT");
        if (!x) die_env("CPULIMIT");
#ifdef RLIMIT_CPU
        doit(RLIMIT_CPU, x);
#endif
        break;
      case '?':
      default:
        die_usage();
        break;
    }
  }

  argv += optind;
  if (!*argv) die_usage();

  pathexec_run(*argv, argv, envp);
  die_exec(*argv);
  return 0;  /* lint */
}
