#include <nemo/stdint.h>
#include <nemo/unixtypes.h>

#include <nemo/uint32.h>
#include <nemo/error.h>
#include <nemo/stralloc.h>
#include <nemo/djbio.h>
#include <nemo/djbiofd.h>
#include <nemo/exit.h>
#include <nemo/str.h>
#include <nemo/scan.h>
#include <nemo/fmt.h>
#include <nemo/macro_unused.h>

/* #include <sys/types.h> */
#include <sys/time.h>
#include <syslog.h>

static char buf[800];  /* syslog truncates long lines (or crashes), GPACIC */
static unsigned int bufpos = 0;  /* 0 <= bufpos < sizeof(buf) */
static unsigned int flag_cont = 0;
static int priority;  /* defined if flag_cont */
static char stamp[FMT_ULONG + FMT_ULONG + 3];  /* defined if flag_cont */

static void stamp_make(void)
{
  struct timeval tv;
  char *s;
  gettimeofday(&tv, (struct timezone *) 0);
  s = stamp;
  s += fmt_ulong(s, (unsigned long) tv.tv_sec);
  *s++ = '.';
  s += fmt_uint0(s, (unsigned int) tv.tv_usec, 6);
  *s = 0;
}

static void flush(void)
{
  if (bufpos) {
    buf[bufpos] = 0;
    if (flag_cont) {
      syslog(priority, "%s+%s", stamp, buf);  /* logger folds invisibly, GPACIC */
    }
    else {
      stamp_make();
      priority = LOG_INFO;
      if (str_start(buf, "warning:")) {
        priority = LOG_WARNING;
      }
      if (str_start(buf, "alert:")) {
        priority = LOG_ALERT;
      }
      syslog(priority, "%s %s", stamp, buf);
      flag_cont = 1;
    }
  }
  bufpos = 0;
}

int main(int argc __UNUSED__, char **argv)
{
  unsigned long facility;
  char ch;

  if (argv[1]) {
    if (argv[2]) {
      scan_ulong(argv[2], &facility);
      openlog(argv[1], 0, (int)(facility << 3));
    }
    else {
      openlog(argv[1], 0, LOG_MAIL);
    }
  }
  else {
    openlog("splogger", 0, LOG_MAIL);
  }

  for (;;) {
    if (djbio_get(djbiofd_in, &ch, 1) < 1) _exit(0);
    if (ch == '\n') {
      flush();
      flag_cont = 0;
      continue;
    }
    if (bufpos == sizeof(buf) - 1) {
      flush();
    }
    if ((ch < 32) || (ch > 126)) {
      ch = '?';  /* logger truncates at 0, GPACIC */
    }
    buf[bufpos++] = ch;
  }
  return 0;
}
