#include <nemo/stdint.h>
#include <nemo/unixtypes.h>
#include <nemo/ndelay.h>
#include <nemo/strerr.h>
#include <nemo/error.h>
#include <nemo/open.h>
#include <nemo/stralloc.h>
#include <nemo/sgetopt.h>
#include <nemo/sig.h>
#include <nemo/unix.h>
#include <nemo/exit.h>

#include "die.h"

const char USAGE[] = "options: u up, d down, o once, x exit, p pause, c continue, h hup, a alarm, i interrupt, t term, k kill";

static const char WARNING[] = "svctl: warning: ";

static stralloc options = STRALLOC;

static int fd_orig_dir;

int main(int argc, char **argv)
{
  char ch;
  int opt;
  int fd;
  const char *dir;

  PROGRAM = argv[0];
  sig_ignore(sig_pipe);

  while ((opt = getopt(argc, argv, "udopchaitkx")) != opteof) {
    if (opt == '?') die_usage();
    ch = (char)opt;
    if (stralloc_chr(&options, ch) != options.len) continue;  /* duplicate */
    if (!stralloc_append(&options, &ch)) die_nomem();
  }
  argv += optind;

  if (!stralloc_len(&options)) die_usage();

  fd_orig_dir = open_read(".");
  if (fd_orig_dir < 0) die_open("current directory");

  while ((dir = *argv++)) {
    if (chdir(dir) < 0) die_chdir(dir);
    fd = open_write("supervise/control");
    if (fd < 0) {
      if (errno == error_nodevice) {
	strerr_warn4x(WARNING, "unable to control ", dir, ": supervise not running");
      }
      else {
	strerr_warn3sys(WARNING, "unable to control ", dir);
      }
    }
    else {
      ndelay_off(fd);
      if (write(fd, options.s, options.len) < 0) {
	strerr_warn3sys(WARNING, "error writing commands to ", dir);
      }
      close(fd);
    }
    if (fchdir(fd_orig_dir) < 0) die_sys("unable to set directory");
  }

  _exit(0);
}
