#include <nemo/stdint.h>
#include <nemo/unixtypes.h>

#include <nemo/strerr.h>
#include <nemo/error.h>
#include <nemo/open.h>
#include <nemo/fmt.h>
#include <nemo/tai.h>
#include <nemo/stralloc.h>
#include <nemo/djbio.h>
#include <nemo/djbiofd.h>
#include <nemo/unix.h>
#include <nemo/exit.h>

#include <sys/stat.h>

#include "die.h"

static const char WARNING[] = "svstat: warning: ";

static byte_t status[18];
static char strnum[FMT_ULONG];

static pid_t pid;
static unsigned int normally_up;
static unsigned int paused;
static char want;

static void doit(const char *dir)
{
  struct stat st;
  ssize_t r;
  int fd;
  const char *x;
  struct tai when;
  struct tai now;

  djbio_puts(djbiofd_out, dir);
  djbio_puts(djbiofd_out, ": ");

  if (chdir(dir) < 0) {
    x = error_str(errno);
    djbio_puts(djbiofd_out, "unable to chdir: ");
    djbio_puts(djbiofd_out, x);
    return;
  }

  normally_up = 0;
  if (stat("down", &st) < 0) {
    if (errno != error_noent) {
      x = error_str(errno);
      djbio_puts(djbiofd_out, "unable to stat down: ");
      djbio_puts(djbiofd_out, x);
      return;
    }
    normally_up = 1;
  }

  fd = open_write("supervise/ok");
  if (fd < 0) {
    if (errno == error_nodevice) {
      djbio_puts(djbiofd_out, "supervise not running");
      return;
    }
    x = error_str(errno);
    djbio_puts(djbiofd_out, "unable to open supervise/ok: ");
    djbio_puts(djbiofd_out, x);
    return;
  }
  close(fd);

  fd = open_read("supervise/status");
  if (fd < 0) {
    x = error_str(errno);
    djbio_puts(djbiofd_out, "unable to open supervise/status: ");
    djbio_puts(djbiofd_out, x);
    return;
  }
  r = read(fd, status, sizeof(status));
  close(fd);
  if (r < (int)sizeof(status)) {
    x = (r < 0) ? error_str(errno) : "bad format";
    djbio_puts(djbiofd_out, "unable to read supervise/status: ");
    djbio_puts(djbiofd_out, x);
    return;
  }

  pid = status[15];

  pid <<= 8;
  pid += status[14];

  pid <<= 8;
  pid += status[13];

  pid <<= 8;
  pid += status[12];

  paused = (unsigned int)status[16];
  want = (char)status[17];

  tai_unpack(&when, status);
  tai_now(&now);
  if (tai_less(&now, &when)) when = now;
  tai_sub(&when, &now, &when);

  if (pid) {
    djbio_puts(djbiofd_out, "up (pid ");
    djbio_put(djbiofd_out, strnum, fmt_ulong(strnum, (unsigned long)pid));
    djbio_puts(djbiofd_out, ") ");
  }
  else {
    djbio_puts(djbiofd_out, "down ");
  }

  djbio_put(djbiofd_out, strnum, fmt_ulong(strnum, (unsigned int)tai_seconds(&when)));
  djbio_puts(djbiofd_out, " seconds");

  if (pid && !normally_up) {
    djbio_puts(djbiofd_out, ", normally down");
  }
  if (!pid && normally_up) {
    djbio_puts(djbiofd_out, ", normally up");
  }
  if (pid && paused) {
    djbio_puts(djbiofd_out, ", paused");
  }
  if (!pid && (want == 'u')) {
    djbio_puts(djbiofd_out, ", want up");
  }
  if (pid && (want == 'd')) {
    djbio_puts(djbiofd_out, ", want down");
  }
}

int main(int argc, char **argv)
{
  int fd_orig_dir;

  PROGRAM = argv[0];
  ++argv;
  --argc;

  fd_orig_dir = open_read(".");
  if (fd_orig_dir < 0) die_open("current directory");

  while (argc--) {
    doit(*argv++);
    djbio_putflush(djbiofd_out, "\n", 1);
    if (fchdir(fd_orig_dir) < 0) die_chdir("directory");
  }

  djbio_flush(djbiofd_out);

  _exit(0);
}
