#include <nemo/stdint.h>
#include <nemo/unixtypes.h>

#include <sys/types.h>
#include <time.h>
#include <sys/time.h>

#include <nemo/stralloc.h>
#include <nemo/djbio.h>
#include <nemo/djbiofd.h>
#include <nemo/fmt.h>
#include <nemo/exit.h>
#include <nemo/char.h>
#include <nemo/tai.h>
#include <nemo/caldate.h>
#include <nemo/caltime.h>
#include <nemo/leapsecs.h>
#include <nemo/macro_unused.h>

static char num[FMT_ULONG];
static char datetime[FMT_CALTIME];

static void get(byte_t *ch)
{
  int r;

  r = djbio_GETC(djbiofd_in, ch);
  if (r == 1) return;
  if (r == 0) {
    if (djbio_flush(djbiofd_out) < 0) _exit(111);
    _exit(0);
  }
  _exit(111);
}

static void out(const void *buf, unsigned int len)
{
  if (djbio_put(djbiofd_out, buf, len) < 0) {
    _exit(111);
  }
}

static uint64_t secs;
static unsigned long nanosecs;
static unsigned long u;
static struct tai timestamp;
struct caltime ct;

int main(int argc __UNUSED__, char **argv __UNUSED__)
{
  byte_t ch;

  if (leapsecs_init() < 0) _exit(111);
  for (;;) {
    get(&ch);
    if (ch == '@') {
      secs = 0;
      nanosecs = 0;
      for (;;) {
        get(&ch);
        u = char_hex_table[ch];
        if (u > 15) break;
        secs <<= 4;
        secs += (nanosecs >> 28);
        nanosecs &= 0xfffffff;
        nanosecs <<= 4;
        nanosecs += u;
      }
      timestamp.x = secs;
      caltime_local(&ct, &timestamp, 0, 0);
      out(datetime, caltime_fmt_short(&ct, datetime));
      out(".", 1);
      out(num, fmt_uint0(num, (unsigned int)(nanosecs), 9));
    }
    for (;;) {
      out(&ch, 1);
      if (ch == '\n') break;
      get(&ch);
    }
  }
  if (djbio_flush(djbiofd_out) < 0) _exit(111);
}
